# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

from tavutil.paypalx import AdaptivePayment, PayRequest
from config import (
    PPX_API_USERNAME, PPX_API_PASSWORD, 
    PPX_API_SIGNATURE, PPX_APP_ID
)

api_credentials = {'api_username': PPX_API_USERNAME,
                    'api_password': PPX_API_PASSWORD,
                    'api_signature': PPX_API_SIGNATURE,
                    'app_id': PPX_APP_ID}
adaptpay = AdaptivePayment(**api_credentials)
pay_request_kwargs = {
            'receiver_email': 'mamadi_1262634965_per@gmail.com',
            'receiver_amount': 20.00, 'currency_code': 'GBP',
            'cancel_url': 'http://www.ebay.com',
            'return_url': 'http://www.sandbox.paypal.com'
}
payreq = PayRequest(**pay_request_kwargs)
payreqmsg = payreq.get_message()
#pp_response = adaptpay.call("PAY", payreqmsg)




# Public Domain (-) 2008-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

r"""
===============
HTML5 Sanitiser
===============

This module provides a single ``sanitise`` function which should protect you
from a wide range of XSS attacks:

  >>> sanitise('<SCRIPT SRC=http://ha.ckers.org/xss.js></SCRIPT>')
  u''

  >>> sanitise("'';!--\"<XSS>=&{()}") == "'';!--\"=&{()}"
  True

  >>> sanitise("<IMG SRC=JaVaScRiPt:alert('XSS')>")
  u'<img />'

  >>> sanitise('<INPUT TYPE="image" SRC="javascript:alert(\'XSS\');">')
  u'<input type="image" />'

  >>> sanitise('<SCRIPT>document.write("<SCRI");</SCRIPT>PT SRC="http://ha.ckers.org/xss.js"></SCRIPT>')
  u'document.write("'

  >>> sanitise('<SCRIPT a=">\'>" SRC="http://ha.ckers.org/xss.js"></SCRIPT>')
  u'\'>" SRC="http://ha.ckers.org/xss.js">'

  >>> sanitise('<<SCRIPT>alert("XSS");//<</SCRIPT>')
  u''

  >>> sanitise('<SCRIPT/SRC="http://ha.ckers.org/xss.js"></SCRIPT>')
  u'SRC="http:/ha.ckers.org/xss.js">'

""" # emacs "'

# See http://ha.ckers.org/xss.html for a listing of various XSS attacks

import re

from BeautifulSoup import BeautifulSoup, CData, Comment, ProcessingInstruction

# ------------------------------------------------------------------------------
# Utility Functions
# ------------------------------------------------------------------------------

create_set = lambda values: frozenset(values.strip().split())
find_cdata = lambda text: isinstance(text, CData)
find_comments = lambda text: isinstance(text, Comment)
find_pi = lambda text: isinstance(text, ProcessingInstruction)

match_js_in_uri_ref = re.compile(
    '(%s)*'
    'j[\s]*(&#x.{1,7})?'
    'a[\s]*(&#x.{1,7})?'
    'v[\s]*(&#x.{1,7})?'
    'a[\s]*(&#x.{1,7})?'
    's[\s]*(&#x.{1,7})?'
    'c[\s]*(&#x.{1,7})?'
    'r[\s]*(&#x.{1,7})?'
    'i[\s]*(&#x.{1,7})?'
    'p[\s]*(&#x.{1,7})?t' % [chr(i) for i in range(33)], re.IGNORECASE
    ).match

match_valid_uri_scheme = re.compile(
    '(http|https|aim|amp|callto|data|dict|dns|fax|fb|feed|file|freenet|ftp|geo|'
    'git|gtalk|im|irc|ircs|itms|lastfm|magnet|mailto|maps|md5|mms|msnim|news|'
    'nntp|psyc|rsync|rtsp|secondlife|sftp|sha|sip|sips|skype|sms|spotify|ssh|'
    'svn|tag|tel|urn|uuid|webcal|xmpp|xri|ymsgr):.*', re.IGNORECASE
    ).match

match_valid_css_value = re.compile(
    '^(#[0-9a-f]+|rgb\(\d+%?,\d*%?,?\d*%?\)?|\d{0,2}\.?\d{0,2}'
    '(cm|deg|em|ex|hz|in|mm|pc|pt|px|s|%|,|\))?)$'
    ).match

# ------------------------------------------------------------------------------
# Some Constants
# ------------------------------------------------------------------------------

VALID_TAGS = create_set("""
    a abbr acronym address area article aside audio b blockquote br button
    canvas caption cite code col colgroup command datalist dd del details dfn
    dialog div dl dt em fieldset figcaption figure footer form h1 h2 h3 h4 h5 h6
    header hgroup hr i img input ins kbd keygen label legend li map mark menu
    meter nav noscript ol optgroup option output p pre progress q rp rt ruby
    samp section select small source span strong sub summary sup table tbody td
    textarea tfoot th thead time tr track ul var video wbr
    """)

# base bdo big body center datagrid dir embed event-source font head html iframe
# link meta object param s script strike style title tt u

VALID_ATTR_PREFIXES = ('aria-', 'data-')

VALID_ATTRS = create_set("""

    class contenteditable contextmenu dir id itemid itemprop itemref itemscope
    itemtype hidden lang role style spellcheck title

    accept accept-charset action alt autocomplete challenge checked cite cols
    colspan coords datetime disabled enctype for form formaction formenctype
    formmethod formnovalidate headers height high href hreflang icon ismap
    keytype kind label list low max maxlength media method min multiple name
    novalidate open optimum pattern ping placeholder preload poster pubdate
    radiogroup readonly rel required reversed rows rowspan scope selected shape
    size span src srclang start step summary type usemap value width wrap

    align border cellpadding cellspacing noshade nowrap

    """)

# accesskey draggable tabindex

# autofocus autoplay controls formtarget loop target

# axis background balance bgcolor bgproperties bordercolor bordercolordark
# bordercolorlight bottompadding ch char charoff charset choff clear color
# compact data datafld datapagesize datasrc default delay dynsrc end face frame
# galleryimg gutter hidefocus hspace inputmode leftspacing longdesc loopcount
# loopend loopstart lowsrc nohref point-size pqg prompt repeat-max repeat-min
# replace rev rightspacing rules scope suppress template toppadding unselectable
# urn valign variable volume vrml vspace

ATTRS_WITH_URI_REFS = frozenset(['action', 'cite', 'formaction', 'href', 'src'])

# background longdesc

VALID_CSS_PROPERTIES = create_set("""

    azimuth background-color border border-bottom border-bottom-color
    border-bottom-left-radius border-bottom-right-radius border-bottom-style
    border-bottom-width border-collapse border-color border-left
    border-left-color border-left-style border-left-width border-radius
    border-right border-right-color border-right-style border-right-width
    border-spacing border-style border-top border-top-color
    border-top-left-radius border-top-right-radius border-top-style
    border-top-width border-width caption-side clear color cursor direction
    display elevation empty-cells float font font-family font-size
    font-size-adjust font-stretch font-style font-variant font-weight height
    ime-mode layout-flow layout-grid layout-grid-char layout-grid-char-spacing
    layout-grid-line layout-grid-mode layout-grid-type letter-spacing
    line-height line-break list-style list-style-position list-style-type marks
    max-height max-width min-height min-width outline-color outline-style
    outline-width overflow overflow-x overflow-y padding padding-bottom
    padding-left padding-right padding-top pause pause-after pause-before pitch
    pitch-range richness ruby-align ruby-overhang ruby-position speak
    speak-header speak-numeral speak-punctuation speech-rate stress table-layout
    text-align text-align-last text-autospace text-decoration text-justify
    text-kashida-space text-overflow text-shadow text-transform unicode-bidi
    vertical-align voice-family volume white-space width word-break word-spacing
    word-wrap writing-mode

    """)

# background* box-shadow* clip column* content counter counter-increment cue*
# filter include-source layer* left/right/top/bottom margin list-style-image
# marker-offset orphans page* play-during position quotes scrollbar* size
# text-indent text-underline-position widows
# z-index zoom

VALID_CSS_KEYWORDS = create_set("""

    left-side far-left left center-left center center-right right
    far-right right-side leftwards rightwards

    below level above lower higher

    alias all-scroll cell col-resize copy count-down count-up count-up-down
    crosshair default grab grabbing hand help move no-drop not-allowed pointer
    progress spinning text vertical-text wait

    caption icon menu message-box small-caption status-bar

    large larger medium small smaller x-large x-small xx-large xx-small

    condensed expanded extra-condensed extra-expanded narrower normal
    semi-condensed semi-expanded ultra-condensed ultra-expanded wider

    active inactive deactivated

    horizontal vertical-ideographic

    both char fixed line loose strict

    inside outside

    armenian circle cjk-ideographic decimal decimal-leading-zero disc georgian
    hebrew katakana-iroha hiragana hiragana-iroha katakana lower-alpha
    lower-greek lower-latin lower-roman square upper-alpha upper-latin
    upper-roman

    crop cross

    medium thick thin

    hidden scroll visible

    high low medium x-high x-low

    above center distribute-letter distribute-space inline left line-edge right
    whitespace

    always code continuous digits once spell-out

    fast medium slow x-fast x-slow

    ideograph-alpha ideograph-numeric ideograph-parenthesis ideograph-space

    distribute distribute-all-lines inter-cluster inter-ideograph inter-word
    newspaper

    clip ellipsis

    capitalize lowercase uppercase

    bidi-override embed

    baseline bottom middle sub super text-bottom text-top top

    collapse hidden hide show visible

    male female child

    loud medium silent soft x-loud x-soft

    break-all break-word keep-all

    lr-tb tb-rl

    arial arial-black comic-sans-ms constantina consolas courier courier-new
    cursive fantasy franklin-gothic-medium geneva georgia helvetica
    helvetica-neue impact lucida-console lucida-grande lucida-sans-unicode
    monaco monospace palatino-linotype sans-serif serif tahoma times
    times-new-roman trebuchet-ms verdana

    !important aqua auto black block blue bold bolder both bottom brown center
    collapse compact dashed dotted double embed fixed fuchsia gray green groove
    hide hidden inherit inline inline-block inset invert italic justify left
    lighter lime line-through list-item ltr marker maroon medium navy none
    normal nowrap oblique olive outset overline pointer pre purple red ridge
    right rtl run-in separate show silver small-caps solid strict table
    table-cell table-column table-column-group table-footer-group
    table-header-group table-row table-row-group teal top transparent underline
    white yellow

    """)

# *-resize (for cursor)
# font-family/size
# text-decoration: blink
# voice-family

VALID_CSS_CLASSES = frozenset()

# ------------------------------------------------------------------------------
# Core Function
# ------------------------------------------------------------------------------

def sanitise(
    html, valid_tags=VALID_TAGS, valid_attrs=VALID_ATTRS,
    valid_attr_prefixes=VALID_ATTR_PREFIXES,
    attrs_with_uri_refs=ATTRS_WITH_URI_REFS,
    valid_css_properties=VALID_CSS_PROPERTIES,
    valid_css_keywords=VALID_CSS_KEYWORDS,
    valid_css_classes=VALID_CSS_CLASSES,
    secure_id_prefix='local-', strip_cdata=True, strip_comments=True,
    strip_pi=True, rel_whitelist=None, second_run=False,
    allow_relative_urls=False, encoding='utf-8'
    ):
    """Return a sanitised version of the provided HTML."""

    soup = BeautifulSoup(html)

    if strip_cdata:
        for cdata in soup.findAll(text=find_cdata):
            cdata.extract()

    if strip_comments:
        for comment in soup.findAll(text=find_comments):
            comment.extract()

    if strip_pi:
        for pi in soup.findAll(text=find_pi):
            pi.extract()

    for tag in soup.findAll(True):

        if tag.name not in valid_tags:
            tag.hidden = True
            continue

        tag_attrs = []; append = tag_attrs.append

        for attr, val in tag.attrs:
            if attr not in valid_attrs:
                for prefix in valid_attr_prefixes:
                    if attr.startswith(prefix):
                        append((attr, val))
                        break
                continue
            if attr == 'id':
                if not val.startswith(secure_id_prefix):
                    continue
            elif attr == 'style':
                new_style = []; add_style = new_style.append
                for style in val.split(u';'):
                    style = style.strip().split(u':', 1)
                    if len(style) != 2:
                        continue
                    prop, value = style
                    prop = prop.strip().lower()
                    if prop not in valid_css_properties:
                        continue
                    components = []; add_component = components.append
                    segments = filter(None, value.split(u','))
                    segcount = len(segments) - 1
                    for idx, segment in enumerate(segments):
                        current = None
                        in_quotes = 0
                        for char in segment:
                            if current is None:
                                if char.isspace():
                                    continue
                                if char == '"' or char == "'":
                                    in_quotes = 1
                                    current = u''
                                else:
                                    current = char
                                continue
                            if in_quotes:
                                if char == '"' or char == "'":
                                    in_quotes = 0
                                    add_component(current)
                                    current = None
                                else:
                                    current += char
                            else:
                                if char.isspace():
                                    add_component(current)
                                    current = None
                                else:
                                    current += char
                        if current:
                            add_component(current)
                        if idx != segcount:
                            add_component(u',')
                    valid = True
                    new_value = []; add_part = new_value.append
                    for component in components:
                        if component == u',':
                            add_part(u', ')
                            continue
                        norm = u'-'.join(component.lower().split())
                        if norm not in valid_css_keywords:
                            if not match_valid_css_value(norm):
                                valid = False
                                break
                        if u' ' in component:
                            add_part(u' "%s"' % component)
                        else:
                            add_part(u' %s' % component)
                    if not valid:
                        continue
                    add_style(u"%s: %s;" % (prop, ''.join(new_value)))
                if not new_style:
                    continue
                val = u''.join(new_style)
            elif attr == 'class':
                val = u' '.join(
                    klass for klass in val.split()
                    if klass in valid_css_classes
                    )
                if not val:
                    continue
            elif attr in attrs_with_uri_refs:
                if allow_relative_urls:
                    if match_js_in_uri_ref(val):
                        continue
                elif not match_valid_uri_scheme(val):
                    continue
            elif (attr == 'rel') and rel_whitelist:
                if val.split(u':')[0] not in rel_whitelist:
                    continue
            append((attr, val))

        tag.attrs = tag_attrs

    # protects against: <<SCRIPT>script>("XSS");//<</SCRIPT>

    if not second_run:
        return sanitise(
            unicode(soup.renderContents(), encoding), valid_tags, valid_attrs,
            valid_attr_prefixes, attrs_with_uri_refs, valid_css_properties,
            valid_css_keywords, valid_css_classes, secure_id_prefix,
            strip_cdata, strip_comments, strip_pi, rel_whitelist, True,
            allow_relative_urls, encoding
            )

    return unicode(soup.renderContents(), encoding)

# ------------------------------------------------------------------------------
# Run Tests
# ------------------------------------------------------------------------------

if __name__ == '__main__':

    import doctest
    doctest.testmod()



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""
================
ZeroConf Support
================

This module provides support functions to register and query ZeroConf records.

The ``register`` function returns either a ``1`` or a ``0`` to indicate a
successful or failed registration.

  >>> register('foo', '_test._tcp', 1234)
  1

And, similarly, the ``query`` function can be used to find registrations for a
given ``regtype``. It takes an optional ``timeout`` value (in seconds) as a
second parameter, e.g.

  >>> query('_test._tcp', 1.0)
  {u'foo._test._tcp.local.': {...'port': 1234...}}

"""

import atexit
import threading

from select import select
from time import time

try:
    import pybonjour
except Exception:
    pybonjour = None

state = threading.local()
state.announce = None
state.query = None
state.current = None
state.timeout = None

def registration_callback(sdRef, flags, errorCode, name, regtype, domain):
    if errorCode == pybonjour.kDNSServiceErr_NoError:
        state.announce = 1
    else:
        state.announce = 0

def register(name, regtype, port):
    if not pybonjour:
        return
    sdRef = pybonjour.DNSServiceRegister(
        name=name, regtype=regtype, port=port, callBack=registration_callback
        )
    try:
        while 1:
            ready = select([sdRef], [], [])
            if sdRef in ready[0]:
                pybonjour.DNSServiceProcessResult(sdRef)
                return state.announce
    finally:
        state.announce = None
        atexit.register(sdRef.close)

def resolve_callback(
    sdRef, flags, interfaceIndex, errorCode, fullname, hosttarget, port,
    txtRecord
    ):
    if errorCode == pybonjour.kDNSServiceErr_NoError:
        record = state.query[fullname] = state.current
        record['host'] = hosttarget
        record['port'] = port

def query_callback(
    sdRef, flags, interfaceIndex, errorCode, serviceName, regtype, replyDomain
    ):
    if errorCode != pybonjour.kDNSServiceErr_NoError:
        return
    if not (flags & pybonjour.kDNSServiceFlagsAdd):
        return
    if state.timeout:
        timeout = state.timeout
    else:
        timeout = None
    state.current = {
        'name': serviceName,
        'type': regtype
        }
    sdRef = pybonjour.DNSServiceResolve(
        0, interfaceIndex, serviceName, regtype, replyDomain, resolve_callback
        )
    try:
        while 1:
            ready = select([sdRef], [], [], timeout)
            if sdRef not in ready[0]:
                break
            return pybonjour.DNSServiceProcessResult(sdRef)
    finally:
        state.current = None
        sdRef.close()

def query(regtype, timeout=5.0):
    if not pybonjour:
        return {}
    sdRef = pybonjour.DNSServiceBrowse(regtype=regtype, callBack=query_callback)
    start = time()
    if timeout:
        state.timeout = timeout
    state.query = {}
    try:
        while (time() - start) <= timeout:
            ready = select([sdRef], [], [], timeout)
            if sdRef in ready[0]:
                pybonjour.DNSServiceProcessResult(sdRef)
        return state.query
    finally:
        state.query = None
        state.timeout = None
        sdRef.close()

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS + doctest.NORMALIZE_WHITESPACE)



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Utility functions for dealing with the operating environment."""

import sys
import subprocess

from os import getcwd

def read_file(filename, mode='rU', content=None):
    """Read and return the contents of the given filename."""

    f = open(filename, mode)
    try:
        content = f.read()
    finally:
        f.close()
    return content

def exit(message, error_code=1):
    """Write an error message to stderr and exit with the given error_code."""
    sys.stderr.write(message + '\n')
    sys.exit(error_code)

class CommandNotFound(Exception):
    """Exception raised when a command line app could not be found."""

def run_command(
    args, retcode=False, reterror=False, exit_on_error=False, error_message="",
    log=None, redirect_stdout=True, redirect_stderr=True, cwd=None,
    shell=sys.platform.startswith('win'), env=None, universal_newlines=True
    ):
    """Execute the command with the given options."""

    log_message = "%s cwd=%s" % (' '.join(args), cwd or getcwd())
    if log:
        if hasattr(log, '__call__'):
            log(log_message)
        else:
            sys.stderr.write("Running command: " + log_message + '\n')

    if redirect_stdout:
        stdout = subprocess.PIPE
    else:
        stdout = None

    if redirect_stderr:
        stderr = subprocess.PIPE
    else:
        stderr = None

    try:
        process = subprocess.Popen(
            args, stdout=stdout, stderr=stderr, shell=shell, cwd=cwd, env=env,
            universal_newlines=universal_newlines
            )
        out, err = process.communicate()
    except OSError:
        error = sys.exc_info()[1]
        if error.errno == 2:
            if not error.filename:
                if exit_on_error:
                    exit("Couldn't find the %r command!" % args[0])
                raise CommandNotFound(args[0])
            raise error
        if exit_on_error:
            exit("Error running: %s\n\n%s" % (log_message, error_message))
        raise

    if process.returncode and exit_on_error:
        if stderr:
            exit_extra = error_message or err
        else:
            exit_extra = error_message or out
        if exit_extra:
            exit("Error running: %s\n\n%s" % (log_message, exit_extra))
        else:
            exit("Error running: %s" % log_message)

    if retcode:
        if reterror:
            return out, err, process.returncode
        return out, process.returncode

    if reterror:
        return out, err
    return out



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

import logging
import socket

from collections import deque
from time import time

from tavutil.async import wrap_method

from tornado.ioloop import IOLoop
from tornado.iostream import IOStream

# ------------------------------------------------------------------------------
# Some Constants
# ------------------------------------------------------------------------------

Loop = IOLoop.instance()

NORMAL_COMMANDS = """
  APPEND AUTH BGREWRITEAOF BGSAVE CONFIG DBSIZE DECR DECRBY EXISTS EXPIRE
  FLUSHALL FLUSHDB GET GETSET HDEL HEXISTS HGET HGETALL HINCRBY HKEYS HLEN HMGET
  HMSET HSET HVALS INCR INCRBY INFO KEYS LASTSAVE LINDEX LLEN LPOP LPUSH LRANGE
  LREM LSET LTRIM MGET MOVE MSET MSETNX PING PSUBSCRIBE PUBLISH PUNSUBSCRIBE
  QUIT RANDOMKEY RENAME RENAMENX RPOP RPOPLPUSH RPUSH SADD SAVE SCARD SDIFF
  SDIFFSTORE SELECT SET SETEX SETNX SHUTDOWN SINTER SINTERSTORE SISMEMBER
  SLAVEOF SMEMBERS SMOVE SORT SPOP SRANDMEMBER SREM SUBSCRIBE SUBSTR SUNION
  SUNIONSTORE TTL TYPE UNSUBSCRIBE ZADD ZCARD ZCOUNT ZINCRBY ZINTERSTORE ZRANGE
  ZRANGEBYSCORE ZRANK ZREM ZREMRANGEBYRANK ZREMRANGEBYSCORE ZREVRANGE ZREVRANK
  ZSCORE ZUNIONSTORE
  """.strip().split()

# ------------------------------------------------------------------------------
# Utility Functions
# ------------------------------------------------------------------------------

def set_max_connections(value):
    Redis._max_cxns = value

# ------------------------------------------------------------------------------
# Exceptions
# ------------------------------------------------------------------------------

class RedisError(Exception):
    pass

# ------------------------------------------------------------------------------
# The Redis Client
# ------------------------------------------------------------------------------

class Redis(object):
    """Async redis client."""

    _global_cxns = {}
    _max_cxns = None
    _open_cxns = 0
    _cxn = None
    _in_txn = 0
    _multi_wait = 0
    _in_progress = 0
    _opened = 0

    def __init__(self, host='', port=6379, unix_socket=''):
        if unix_socket:
            self._addr = addr = unix_socket
        else:
            self._addr = addr = (host, port)
        if addr not in self._global_cxns:
            self._cxns = self._global_cxns[addr] = set()
        else:
            self._cxns = self._global_cxns[addr]

    def handle_connection_close(self, cxn):
        if not cxn._discarded:
            cxn._discarded = 1
            Redis._open_cxns -= 1
            while cxn.queue:
                errback = cxn.queue.popleft()[-1]
                if errback:
                    try:
                        errback(socket.error("Connection closed."))
                    except Exception:
                        pass
            del cxn.queue
        self._cxns.discard(cxn)

    def close_connection(self):
        cxn, self._cxn = self._cxn, None
        if not cxn:
            return
        try:
            cxn.close()
        except socket.error:
            pass

    for _spec in [
        ('SEND_REQUEST', None, '', ''),
        ('DEL', 'delete', '', ''),
        ('BLPOP', """
        if not isinstance(args[-1], (int, float)): args = args + (0,)""", ""),
        ('BRPOP', """
        if not isinstance(args[-1], (int, float)): args = args + (0,)""", ""),
        ('MULTI', '', "txn = 1"),
        ('EXEC', 'execute', '', "multi = txn_end = 1"),
        ('DISCARD', '', "txn_end = 1"),
        ('WATCH', '', "txn = 1"),
        ('UNWATCH', '', "txn = 1"),
        ('MONITOR', '', "persist = 1"),
        ] + [(cmd, '', '') for cmd in NORMAL_COMMANDS]:
        if len(_spec) == 3:
            _command, _before, _after = _spec
            _name = _command.lower()
        else:
            _command, _name, _before, _after = _spec
        if not _name:
            _name = 'send_request'
            _extra_1 = 'cmd, '
            _extra_2 = 'args = (cmd,) + args'
            _extra_3 = 'args[0], '
        else:
            _extra_1 = _extra_3 = ''
            _extra_2 = 'args = (%r,) + args' % _command

        exec(r"""def %s(self, %s*args, **kwargs):

        %s

        if not args:
            self.close_connection()
            raise ValueError("No arguments specified for redis call.")

        cxn = self._cxn
        if not cxn:
            cxns = self._cxns
            max_cxns = Redis._max_cxns
            if (not cxns) and max_cxns and Redis._open_cxns > max_cxns:
                closed = None
                for addr, open_cxns in self._global_cxns.iteritems():
                    if open_cxns:
                        closed = open_cxns.pop()
                        closed.close()
                        break
                if not closed:
                    if '_start' not in kwargs:
                        kwargs['_start'] = time()
                    now = time()
                    if (now - kwargs['_start']) > 60:   # @/@ configurable?
                        kwargs['_start'] = now
                        logging.warn(
                            "Redis connection starving [0x%%x] %%r"
                            %% (id(self), self._addr)
                            )
                    Loop.add_timeout(
                        now + 0.1,
                        lambda: self.%s.__raw__(self, %s *args[1:], **kwargs)
                    )
                    return self
            if cxns:
                cxn = self._cxn = cxns.pop()
            else:
                Redis._open_cxns += 1
                addr = self._addr
                try:
                    if isinstance(addr, str):
                        sock = socket.socket(
                            socket.AF_UNIX, socket.SOCK_STREAM
                        )
                    else:
                        sock = socket.socket(
                            socket.AF_INET, socket.SOCK_STREAM
                        )
                        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                    sock.connect(addr)
                    cxn = IOStream(sock)
                    cxn.queue = deque()
                    cxn._discarded = 0
                except Exception:
                    Redis._open_cxns -= 1
                    raise
                cxn._close_callback = lambda: self.handle_connection_close(cxn)
                self._cxn = cxn

        %s

        request = ['*%%i\r\n' %% len(args)]; out = request.append
        for arg in args:
            if isinstance(arg, unicode):
                arg = arg.encode('utf-8')
            else:
                arg = str(arg)
            out('$%%i' %% len(arg))
            out('\r\n')
            out(arg)
            out('\r\n')

        try:
            cxn.write(''.join(request))
        except socket.error:
            self.close_connection()
            raise

        multi = txn = txn_end = persist = None

        %s

        callback = kwargs.pop('callback', None)
        errback = kwargs.pop('errback', None)
        cxn.queue.append([txn, txn_end, multi, persist, 1, callback, errback])
        self.handle_response()

        return self""" % (_name, _extra_1, _extra_2, _name, _extra_3, _before, _after))

        locals()[_name] = wrap_method(locals()[_name])

    del _command, _name, _before, _after, _spec, _extra_1, _extra_2, _extra_3

    def callback(self, result, err=None):
        self._in_progress = 0
        queue = self._cxn.queue
        if self._multi_wait:
            if err:
                self._multi_wait = None
                del self._multi_results, self._multi_result_left
            elif self._multi_result_left is None:
                self._multi_result_left = left = int(result[1:-2])
                if left == -1:
                    result = None
                    self._multi_wait = None
                    del self._multi_results, self._multi_result_left
                else:
                    return self._cxn.read_until('\r\n', self.handle_response)
            else:
                self._multi_result_left -= 1
                self._multi_results.append(result)
                if self._multi_result_left:
                    return self._cxn.read_until('\r\n', self.handle_response)
                else:
                    result = self._multi_results
                    self._multi_wait = None
                    del self._multi_results, self._multi_result_left
        txn, txn_end, multi, persist, stage, callback, errback = queue.popleft()
        if self._in_txn and txn_end:
            self._in_txn = 0
        cb = errback if err else callback
        Redis._opened += 1
        if cb:
            cb(result)
        if persist:
            queue.appendleft([txn, txn_end, multi, persist, 1, callback, errback])
        if queue:
            Loop.add_callback(self.handle_response)
        else:
            if not self._in_txn:
                self._cxns.add(self._cxn)
                self._cxn = None

    def errback(self, error):
        self.callback(error, 1)

    def handle_response(self, data=None):
        try:
            cxn = self._cxn
            txn, _, multi, _, stage, _, _ = cxn.queue[0]
            if data is None:
                if self._in_progress:
                    return
                if txn:
                    self._in_txn = 1
                self._in_progress = 1
                if multi:
                    self._multi_wait = 1
                    self._multi_results = []
                    self._multi_result_left = None
                    return cxn.read_until('\r\n', self.callback)
                return cxn.read_until('\r\n', self.handle_response)
            if stage == 1:
                opener = data[0]
                if opener == '+':
                    return self.callback(data[1:-2])
                if opener == ':':
                    return self.callback(int(data[1:-2]))
                if opener == '$':
                    length = int(data[1:-2])
                    if length == -1:
                        return self.callback(None)
                    cxn.queue[0][-3] = 2
                    return cxn.read_bytes(length+2, self.handle_response)
                if opener == '-':
                    return self.errback(RedisError(data[:-2]))
                if opener == '*':
                    self._results = []
                    self._result_left = int(data[1:-2])
                    cxn.queue[0][-3] = 3
                    return cxn.read_until('\r\n', self.handle_response)
                return self.errback(RedisError("Unknown response %r" % data))
            if stage == 2:
                return self.callback(data[:-2])
            if stage == 3:
                length = int(data[1:-2])
                if length == -1:
                    self._results.append(None)
                    self._result_left -= 1
                    if self._result_left:
                        return cxn.read_until('\r\n', self.handle_response)
                else:
                    cxn.queue[0][-3] = 4
                    return cxn.read_bytes(length+2, self.handle_response)
            if stage == 4:
                self._results.append(data[:-2])
                self._result_left -= 1
                if self._result_left:
                    cxn.queue[0][-3] = 3
                    return cxn.read_until('\r\n', self.handle_response)
            results = self._results
            del self._results, self._result_left
            return self.callback(results)
        except Exception:
            self.close_connection()
            raise


if __name__ == '__main__':

    from tavutil.async import async

    def constructor():
        return Redis(unix_socket='/tmp/redis.sock')

    redis = constructor()
    def handle_get(result):
        print "GOT:", result

    redis.set('name', 'tav', run=1)
    redis.get('name')(handle_get)

    N = 200
    set_max_connections(5)

    @async
    def test_set(i):
        redis = constructor()
        x = yield redis.send_request('set', 'foo', i)
        yield redis.multi()
        x = yield redis.set('foo', i)
        print x
        try:
            x = yield redis.get('foo', i)
        except Exception, error:
            print "aaas", error
        x = yield redis.incr('foos')
        print x
        x = yield redis.incr('foo')
        print repr(x)
        x = yield redis.execute()
        print repr(x)
        return
        try:
            x = yield redis.send_request('ping')
        except Exception, error:
            print 'got error', repr(error), "fa"
            return
        # x = yield redis.ping()
        print x
        #_ = yield redis.set('foo', i)
        # print _, i

    @async
    def test_get():
        redis = constructor()
        result = yield redis.send_request('get', 'foo')
        print "Got:", result

    def test_monitor():
        def handle_monitor_line(line):
            print "mon", line
        redis = constructor()
        redis.monitor()(handle_monitor_line)

    test_monitor()

    for i in xrange(N):
        test_set(i)
    else:
        print 'get'
        test_get()

    def print_max():
        print Redis._opened
        Loop.add_timeout(time() + 1, print_max)

    print_max()

    def on_event(result):
        print "GOT!!", result

    r = constructor()
    r.blpop('hmz')(callback=on_event)

    Loop.start()

# r = Redis('espians.com', 9094)

# def foo(i):
#     r.set('foo', i)
#     r.close_connection()

# from gevent import joinall
# from time import time

# start = time()
# joinall(
#     [spawn(foo, i) for i in range(1000)]
#     )
# print time() - start
#print r.get('foo')

# r.set('foo', 'bar')



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""
===================
IPv4/IPv6 Addresses
===================

This module provides a lightweight library for manipulating IP addresses:

  >>> IPv4('123.45.67.89')
  123.45.67.89

  >>> IPv4('123.45.67.89') == IPv4('123.45.67.89')
  True

  >>> IPv4('123.45.67.89') > IPv4('123.45.67.88')
  True

  >>> IPv6('::1')
  ::1

  >>> IPv6('0:0::0:1')
  ::1

  >>> IPv6('::')
  ::

  >>> IPv6('::ffff:123.45.67.89')
  ::ffff:7b2d:4359

The ``get_teredo_ipv4_address`` method unobfuscates the public IPv4 address
embedded within Teredo IPv6 addresses:

  >>> ip = IPv6('2001::53aa:64c:0:3ffe:a3e2:b3d8')
  >>> ip.get_teredo_ipv4_address()
  92.29.76.39

The method returns ``None`` for non-Teredo addresses:

  >>> ip = IPv6('::1')
  >>> ip.get_teredo_ipv4_address()

"""

from ipaddr import IPv6Address as UtilityClass

# ------------------------------------------------------------------------------
# Helper Objects
# ------------------------------------------------------------------------------

compress_hextets = UtilityClass(0)._compress_hextets

class InvalidIPAddress(ValueError):
    pass

# ------------------------------------------------------------------------------
# IPv4 Address
# ------------------------------------------------------------------------------

class IPv4(long):

    ip_str = None
    max_ip = (2 ** 32) - 1

    def __new__(klass, ip, int_types=(int, long)):
        if isinstance(ip, int_types):
            if not 0 <= ip <= klass.max_ip:
                raise InvalidIPAddress(ip)
            return super(IPv4, klass).__new__(klass, ip)
        try:
            octets = ip.split('.')
        except Exception:
            raise InvalidIPAddress(ip)
        if len(octets) != 4:
            raise InvalidIPAddress(ip)
        ip_val = 0
        for octet in octets:
            octet = int(octet, 10)
            if not 0 <= octet <= 255:
                raise InvalidIPAddress(ip)
            ip_val = (ip_val << 8) | octet
        self = super(IPv4, klass).__new__(klass, ip_val)
        self.ip_str = ip
        return self

    def __repr__(self):
        if not self.ip_str:
            ip_val = self
            octets = []; out = octets.append
            for i in range(4):
                out(str(ip_val & 255))
                ip_val >>= 8
            self.ip_str = '.'.join(octets[::-1])
        return self.ip_str

# ------------------------------------------------------------------------------
# IPv6 Address
# ------------------------------------------------------------------------------

class IPv6(long):

    ip_str = None
    max_ip = (2 ** 128) - 1

    def __new__(klass, ip, int_types=(int, long)):

        if isinstance(ip, int_types):
            if not 0 <= ip <= klass.max_ip:
                raise InvalidIPAddress(ip)
            return super(IPv6, klass).__new__(klass, ip)

        if not isinstance(ip, basestring):
            raise InvalidIPAddress(ip)

        if ip.count('.') == 3:
            hextets = ip.split(':')
            try:
                ipv4 = IPv4(hextets.pop())
            except ValueError:
                raise InvalidIPAddress(ip)
            hextets.extend([
                '%04x' % ((ipv4 >> 16) & 65535),
                '%04x' % (ipv4 & 65535),
                ])
            ip = ':'.join(hextets)

        if ip.count('::'):
            head, tail = ip.split('::')
            ip = [
                '%04x' % (int(hextet, 16) if hextet else 0)
                for hextet in head.split(':')
                ]
            ip += ['0000'] * (8 - len(head.split(':')) - len(tail.split(':')))
            ip += [
                '%04x' % (int(hextet, 16) if hextet else 0)
                for hextet in tail.split(':')
                ]
            ip = ':'.join(ip)

        hextets = ip.split(':')
        if len(hextets) != 8:
            raise InvalidIPAddress(ip)

        ip_val = 0
        for hextet in hextets:
            hextet = int(hextet, 16)
            if not 0 <= hextet <= 65535:
                raise InvalidIPAddress(ip)
            ip_val = (ip_val << 16) | hextet

        return super(IPv6, klass).__new__(klass, ip_val)

    def __repr__(self):
        if not self.ip_str:
            hex_str = '%032x' % self
            hextets = ['%x' % int(hex_str[x:x+4], 16) for x in range(0, 32, 4)]
            self.ip_str = ':'.join(compress_hextets(hextets))
        return self.ip_str

    def get_teredo_ipv4_address(self, prefix='20010000'):
        ip = '%032x' % self
        if ip[:8] != prefix:
            return
        return IPv4(int('0x' + ip[-8:], 16) ^ 0xffffffff)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Utility functions to help with detecting the SCM system being used."""

from os.path import abspath, isdir
from tavutil.env import run_command, CommandNotFound


def is_git():
    """Return whether the current directory is inside a Git repo."""

    try:
        _, error = run_command(
            ["git", "rev-parse", "--is-inside-work-tree"], retcode=True
            )
    except CommandNotFound:
        return

    if not error:
        return True


def is_mercurial():
    """Return whether the current directory is inside a Mercurial repo."""

    try:
        _, error = run_command(["hg", "root"], retcode=True)
    except CommandNotFound:
        return

    if not error:
        return True


def is_subversion():
    """Return whether the current directory is inside a Subversion repo."""

    if isdir('.svn'):
        return True


def guess(priority='git'):
    """Tries to guess the SCM being used in the current directory."""

    if priority == 'git':
        if is_git():
            return 'git'
        if is_mercurial():
            return 'hg'
        if is_subversion():
            return 'svn'
    elif priority == 'hg':
        if is_mercurial():
            return 'hg'
        if is_git():
            return 'git'
        if is_subversion():
            return 'svn'
    elif priority == 'svn':
        if is_subversion():
            return 'svn'
        if is_git():
            return 'git'
        if is_mercurial():
            return 'hg'
    else:
        raise ValueError("Unknown SCM passed as the priority: %r" % priority)


class SCMBase(object):
    """SCM handler base class."""

    def __init__(self, preferred_scm='git'):
        self._preferred_scm = preferred_scm
        self._scm = None
        self._root = None

    @property
    def scm(self):
        if not self._scm:
            scm = guess(self._preferred_scm)
            if scm not in ['git']:
                raise NotImplementedError(
                    "Sorry, support not yet implemented for: %r" % scm
                    )
            self._scm = scm
        return self._scm

    @property
    def root(self):
        if not self._root:
            if self.scm == 'git':
                self._root = abspath(
                    run_command(['git', 'rev-parse', '--show-cdup']).strip()
                    )
        return self._root


class SCMConfig(SCMBase):
    """SCM Configuration handler."""

    def __init__(self, preferred_scm='git'):
        super(SCMConfig, self).__init__(preferred_scm)
        self._config_cache = {}

    def get(self, prop, default=None):
        if prop in self._config_cache:
            return self._config_cache[prop]
        if self.scm == 'git':
            value, error = run_command(['git', 'config', prop], retcode=True)
            if error:
                value = default
            else:
                value = value.strip()
        return self._config_cache.setdefault(prop, value)

    def set(self, prop, value):
        if self.scm == 'git':
            _, error = run_command(
                ['git', 'config', prop, value], retcode=True
                )
            if error:
                raise IOError("Couldn't set: git config %s %s" % (prop, value))
        self._config_cache[prop] = value

    def delete(self, prop, value_regex=None, section=None, all=None):
        if self.scm == 'git':
            if all:
                args = ['--unset-all', prop]
                if value_regex:
                    args.append(value_regex)
            elif section:
                args = ['--remove-section', prop]
            else:
                args = ['--unset', prop]
                if value_regex:
                    args.append(value_regex)
            run_command(['git', 'config'] + args)
            self._config_cache.clear()


if __name__ == '__main__':
    print guess('hg')



# Changes to this file by The Tavutil Authors are in the Public Domain.
# See the Tavutil UNLICENSE file for details.

# Adapted from http://code.activestate.com/recipes/277940/
# Thanks Raymond Hettinger and Paul Cannon!

"""
===========================
Automatic Code Optimisation
===========================

When hand optimising functions, it is typical to include lines like:

    >>> def some_function():
    ...
    ...     _len = len
    ...     _int = int
    ...
    ...     # do something with _len and _int ...

This is useful in functions with tight loops. By assigning to variables in the
local namespace, code can avoid wasting time doing repeated global namespace
lookups.

This module provides an optimiser which binds variables like ``len`` to local
constants. The speed-up is substantial -- trading one or more dictionary-based
namespace lookups for a single C-speed array lookup -- and keeps the original
code free of clutter like _len=len.

You use ``optimise`` like a normal decorator:

    >>> import random

    >>> @optimise(verbose=True)
    ... def sample(population, k):
    ...     '''Return k unique random elements from a population sequence.'''
    ...
    ...     if not isinstance(population, (list, tuple, str)):
    ...         raise TypeError('Cannot handle type', type(population))
    ...
    ...     n = len(population)
    ...
    ...     if not 0 <= k <= n:
    ...         raise ValueError('Sample larger than population')
    ...
    ...     result = [None] * k
    ...     pool = list(population)
    ...
    ...     for i in xrange(k):         # invariant:  non-selected at [0,n-i)
    ...         j = int(random.random() * (n-i))
    ...         result[i] = pool[j]
    ...         pool[j] = pool[n-i-1]   # move non-selected item into vacancy
    ...
    ...     return result
    isinstance --> <built-in function isinstance>
    list --> <type 'list'>
    tuple --> <type 'tuple'>
    str --> <type 'str'>
    TypeError --> <type 'exceptions.TypeError'>
    type --> <type 'type'>
    len --> <built-in function len>
    ValueError --> <type 'exceptions.ValueError'>
    list --> <type 'list'>
    xrange --> <type 'xrange'>
    int --> <type 'int'>
    random --> <module 'random' from '...'>
    new folded constant: (<type 'list'>, <type 'tuple'>, <type 'str'>)
    new folded constant: <built-in method random of Random object at ...>

Variables like ``random`` from the global namespace and ``len``, ``ValueError``,
``list``, ``xrange``, and ``int`` from the builtin namespace are replaced with
constants. If ``random`` had not already been imported at binding time, then it
would have been left as a runtime global lookup.

Constant binding is especially useful for:

* Functions with many references to global constants (sre_compile for example).
* Functions relying on builtins (such as len(), True, and False).
* Functions that raise or catch exceptions (such as IndexError).
* Recursive functions (to optimise the function's reference to itself).
* Functions that call other functions defined in the same module.
* Functions accessing names imported into the global namespace (i.e. from random
  import random).

After binding globals to constants, the decorator makes a second pass and folds
constant attribute lookups and tuples of constants into single constants. The
first frequently occurs with module references (such as string.hexdigits). The
second commonly occurs in for-loops and conditionals using the ``in`` operator
such as: for x in ('a','b','c').

Binding should be applied selectively to those functions where speed is
important and dynamic updates are not desired (i.e. the globals do not change).

In more dynamic environments, a more conservative approach is to set
builtin_only to True so that only the builtins get optimised (this includes
functions like len(), exceptions like IndexError, and constants like True or
False). Alternatively, individual variable names can be added to a stoplist so
that the function knows to leave them unchanged.

Sine optimise() works like other function decorators such as classmethod() and
staticmethod(), it can be used inside a class definition to optimise methods.

When optimising recursive functions, you need to use the following approach
instead of using the @optimise decorator::

    sample = _make_constants(sample, verbose=True)

A utility ``optimise_all`` function is also provided to automatically decorate
every function and method in a module or class. Call optimise_all(my_class)
after the class definition or put optimise_all(sys.modules[__name__]) as the
last line in a module needing optimisation.

"""

import __builtin__
import logging
import sys

from opcode import opmap, HAVE_ARGUMENT, EXTENDED_ARG
from types import FunctionType, ClassType, CodeType

# ------------------------------------------------------------------------------
# some konstants
# ------------------------------------------------------------------------------

globals().update(opmap)

def _make_constants(
    function, builtins=None, builtin_only=False, stoplist=(), constant_fold=True,
    verbose=False,
    ):
    """Return the given ``function`` with its constants folded."""

    if verbose == 2:
        logging.info(
            "# OPTIMISING : %s.%s", function.__module__, function.__name__
            )

    co = function.func_code
    newcode = map(ord, co.co_code)
    newconsts = list(co.co_consts)
    names = co.co_names
    codelen = len(newcode)

    if builtins:
        env = vars(builtins).copy()
    else:
        env = vars(__builtin__).copy()

    if builtin_only:
        stoplist = dict.fromkeys(stoplist)
        stoplist.update(function.func_globals)
    else:
        env.update(function.func_globals)

    # first pass converts global lookups into constants

    i = 0

    while i < codelen:
        opcode = newcode[i]
        if opcode in (EXTENDED_ARG, STORE_GLOBAL):
            # for simplicity, only optimise common cases
            logging.info("\n\nFound opcode\n\n")
            return function
        if opcode == LOAD_GLOBAL:
            oparg = newcode[i+1] + (newcode[i+2] << 8)
            name = co.co_names[oparg]
            if name in env and name not in stoplist:
                value = env[name]
                for pos, v in enumerate(newconsts):
                    if v is value:
                        break
                else:
                    pos = len(newconsts)
                    newconsts.append(value)
                newcode[i] = LOAD_CONST
                newcode[i+1] = pos & 0xFF
                newcode[i+2] = pos >> 8
                if verbose:
                    logging.info("%s --> %s", name, value)
        i += 1
        if opcode >= HAVE_ARGUMENT:
            i += 2

    # second pass folds tuples of constants and constant attribute lookups

    if constant_fold:

        i = 0

        while i < codelen:

            newtuple = []
            while newcode[i] == LOAD_CONST:
                oparg = newcode[i+1] + (newcode[i+2] << 8)
                newtuple.append(newconsts[oparg])
                i += 3

            opcode = newcode[i]
            if not newtuple:
                i += 1
                if opcode >= HAVE_ARGUMENT:
                    i += 2
                continue

            if opcode == LOAD_ATTR:
                obj = newtuple[-1]
                oparg = newcode[i+1] + (newcode[i+2] << 8)
                name = names[oparg]
                try:
                    value = getattr(obj, name)
                except AttributeError:
                    continue
                deletions = 1

            elif opcode == BUILD_TUPLE:
                oparg = newcode[i+1] + (newcode[i+2] << 8)
                if oparg != len(newtuple):
                    continue
                deletions = len(newtuple)
                value = tuple(newtuple)

            else:
                continue

            reljump = deletions * 3
            newcode[i-reljump] = JUMP_FORWARD
            newcode[i-reljump+1] = (reljump-3) & 0xFF
            newcode[i-reljump+2] = (reljump-3) >> 8

            n = len(newconsts)
            newconsts.append(value)
            newcode[i] = LOAD_CONST
            newcode[i+1] = n & 0xFF
            newcode[i+2] = n >> 8

            i += 3

            if verbose:
                logging.info("New folded constant: %s", value)

    codestr = ''.join(map(chr, newcode))
    codeobj = CodeType(
        co.co_argcount, co.co_nlocals, co.co_stacksize,
        co.co_flags, codestr, tuple(newconsts), co.co_names,
        co.co_varnames, co.co_filename, co.co_name,
        co.co_firstlineno, co.co_lnotab, co.co_freevars,
        co.co_cellvars
        )

    return FunctionType(
        codeobj, function.func_globals, function.func_name,
        function.func_defaults, function.func_closure
        )

_make_constants = _make_constants(_make_constants) # optimise thyself!

def optimise_all(
    module_or_class, builtins=None, builtin_only=False, stoplist=(),
    constant_fold=True, verbose=False
    ):
    """Recursively apply constant binding to functions in a module or class."""

    try:
        d = vars(module_or_class)
    except TypeError:
        return

    for k, v in d.items():
        if type(v) is FunctionType:
            new_v = _make_constants(
                v, builtins, builtin_only, stoplist, constant_fold, verbose
                )
            setattr(module_or_class, k, new_v)
        elif type(v) in (type, ClassType):
            optimise_all(
                v, builtins, builtin_only, stoplist, constant_fold, verbose
                )

optimise_all = _make_constants(optimise_all)

@_make_constants
def optimise(
    builtins=None, builtin_only=False, stoplist=(), constant_fold=True,
    verbose=False
    ):
    """Return a decorator for optimising global references."""

    if type(builtins) is FunctionType:
        raise ValueError("The optimise decorator must have arguments.")

    return lambda f: _make_constants(
        f, builtins, builtin_only, stoplist, constant_fold, verbose
        )

def build_optimising_metaclass(
    builtins=None, builtin_only=False, stoplist=(), constant_fold=True,
    verbose=False
    ):
    """Return a automatically optimising metaclass for use as __metaclass__."""

    class _OptimisingMetaclass(type):
        def __init__(cls, name, bases, dict):
            super(_OptimisingMetaclass, cls).__init__(name, bases, dict)
            optimise_all(
                cls, builtins, builtin_only, stoplist, constant_fold, verbose
                )

    return _OptimisingMetaclass

OptimisingMetaclass = build_optimising_metaclass()



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Miscellaneous utility crypto functions."""

from base64 import b64encode, b32encode
from hmac2 import HMAC
from hashlib import sha384
from os import urandom
from time import time
from urllib import urlencode

# ------------------------------------------------------------------------------
# http://rdist.root.org/2009/05/28/timing-attack-in-google-keyczar-library/
# ------------------------------------------------------------------------------

def secure_string_comparison(s1, s2, ord=ord):
    """Securely compare 2 strings in a manner which avoids timing attacks."""

    if len(s1) != len(s2):
        return False

    total = 0

    for x, y in zip(s1, s2):
        total |= ord(x) ^ ord(y)

    return total == 0

# ------------------------------------------------------------------------------
# tamper-proof value generation
# ------------------------------------------------------------------------------

def create_encoded_mac(value, key, hmac=HMAC, hasher=sha384):
    """Return a base64-encoded MAC."""

    digest = hmac(key, value, hasher).digest()
    return b64encode(digest, '-_').rstrip('=')


def create_tamper_proof_string(
    name, value, key, duration=None, hmac=HMAC, hasher=sha384
    ):
    """Return a tamper proof version of the passed in string value."""

    if not isinstance(name, basestring):
        raise ValueError("You can only tamper-proof str name/values.")

    if not isinstance(value, basestring):
        raise ValueError("You can only tamper-proof str name/values.")

    if duration:
        value = "%s:%s" % (int(time()) + duration, value)

    named_value = "%s|%s" % (name.replace('|', r'\|'), value)

    return "%s:%s" % (
        create_encoded_mac(named_value, key, hmac, hasher),
        value
        )


def validate_tamper_proof_string(
    name, value, key, timestamped=False, hmac=HMAC, hasher=sha384
    ):
    """Validate that the given value hasn't been tampered with."""

    if not isinstance(name, basestring):
        raise ValueError("You can only tamper-proof str name/values.")

    if not isinstance(value, basestring):
        raise ValueError("You can only tamper-proof str name/values.")

    try:
        mac, value = value.split(':', 1)
    except:
        return

    named_value = "%s|%s" % (name.replace('|', r'\|'), value)

    expected_mac = create_encoded_mac(named_value, key, hmac, hasher)
    if not secure_string_comparison(mac, expected_mac):
        return

    if timestamped:
        try:
            timestamp, value = value.split(':', 1)
            timestamp = int(timestamp)
        except:
            return
        if time() > timestamp:
            return

    return value

# ------------------------------------------------------------------------------
# pseudo-signature generation and validation
# ------------------------------------------------------------------------------

def create_signature_for_payload(payload, key):
    """Return the signature for the given payload."""

    payload_keys = sorted(payload.keys())
    output = []; append = output.append

    for payload_key in payload_keys:
        append(urlencode({payload_key: payload[payload_key]}))

    output = '&'.join(output)
    return create_encoded_mac(output, key)


def sign_payload(payload, key, nonce_name='nonce', nonce_size=20):
    """Return a signature and a modified payload with a generated nonce."""
    payload[nonce_name] = b32encode(urandom(nonce_size))
    return create_signature_for_payload(payload, key), payload


def validate_signed_payload(payload, key, signature):
    """Validate that the signature matches the given payload and key."""
    expected_sig = create_signature_for_payload(payload, key)
    if secure_string_comparison(signature, expected_sig):
        return payload



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Service input validation support."""

from time import time
from sha import new as sha1

# ------------------------------------------------------------------------------
# some konstants
# ------------------------------------------------------------------------------

CO_VARARGS = 4
CO_VARKEYWORDS = 8

# ------------------------------------------------------------------------------
# kore validate funktion
# ------------------------------------------------------------------------------

def validate(**spec):
    """Validation decorator which enforces the specified ``spec``."""

    def __decorate(func):

        rkey = sha1(str(time())).hexdigest()[:7]
        rkw = 'kw_%s' % rkey

        code = func.func_code
        varnames = list(code.co_varnames)
        defaults = func.func_defaults or ()
        func_name = func.func_name

        varargs = varkwargs = None

        if code.co_flags & CO_VARKEYWORDS:
            varkwargs = varnames.pop(-1)

        if code.co_flags & CO_VARARGS:
            varargs = varnames.pop(-1)

        params = [varnames.pop(0)]; add = params.append
        params2 = params[:]; add2 = params2.append
        default_pointer = len(varnames) - len(defaults)

        for idx, varname in enumerate(varnames):
            if idx < default_pointer:
                add(varname)
                add2("%s[%r]" % (rkw, varname))
            else:
                add("%s=defaults_%s[%s]" % (varname, rkey, idx-default_pointer))
                add2("%s=%s.get(%r, defaults_%s[%s])" % (varname, rkw, varname, rkey, idx-default_pointer))

        if varargs:
            add("*%s" % varargs)
            add2("*%s[%r]" % (rkw, varargs))
        if varkwargs:
            add("**%s" % varkwargs)
            add2("**%s[%r]" % (rkw, varkwargs))
            
        params = ", ".join(params)
        params2 = ", ".join(params2)

        source = """
def %(func_name)s(%(params)s):
    kws_%(r)s = locals()
    kw_%(r)s = {}
    for key_%(r)s in kws_%(r)s:
        if key_%(r)s in spec_%(r)s: 
            try:
                kw_%(r)s[key_%(r)s] = spec_%(r)s[key_%(r)s](kws_%(r)s[key_%(r)s])
            except Exception:
                raise ValueError(
                    "Could not validate input argument '%%s=%%s'" %% (key_%(r)s, kws_%(r)s[key_%(r)s])
                    )
        else:
            kw_%(r)s[key_%(r)s] = kws_%(r)s[key_%(r)s]
    return func_%(r)s(%(params2)s)
""" % dict(params=params, params2=params2, r=rkey, func_name=func_name)

        env = {
            'func_%s' % rkey: func,
            'spec_%s' % rkey: spec,
            'defaults_%s' % rkey: defaults,
            'ValueError': ValueError
            }
        exec source in env

        return env[func_name]

    return __decorate



# Public Domain (-) 2004-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""I/O Helper Utilities."""

import sys

from textwrap import TextWrapper

try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO

# ------------------------------------------------------------------------------
# Some Constants
# ------------------------------------------------------------------------------

# Colours -- dedicated to Saritah who loves pretty colours.
# The beautiful colours!

CODES = {

        'reset':'\x1b[0m',
        'black':'\x1b[30m',
         'blue':'\x1b[34;01m',
         'bold':'\x1b[01m',    # will differ in various terms
        'brown':'\x1b[33;06m',
     'darkblue':'\x1b[34;06m',
    'darkgreen':'\x1b[32;06m',
     'darkgrey':'\x1b[30;01m',
      'darkred':'\x1b[31;06m',
        'green':'\x1b[32;01m',
         'grey':'\x1b[37;06m',
      'magenta':'\x1b[35;01m',
       'purple':'\x1b[35;06m',
          'red':'\x1b[31;01m',
         'teal':'\x1b[36;06m',
    'turquoise':'\x1b[36;01m',
       'yellow':'\x1b[33;01m',
        'white':'\x1b[37;01m',
         'null':'',

    }

if sys.platform == 'win32':
    from collections import defaultdict
    CODES = defaultdict(str)

# ------------------------------------------------------------------------------
# Terminal Formatters
# ------------------------------------------------------------------------------

def wrap_text(text, maxwidth=80, leadpadding=0):
    """Wrap the given ``text`` into lines of specific length."""
    wrap = TextWrapper(width=maxwidth)
    if leadpadding:
        wrap.subsequent_indent = ' ' * leadpadding
    return wrap.fill(text)

def print_text(colour, text):
    """Return text in the specified colour."""
    return CODES[colour] + text + CODES['reset']

def print_reset():
    return CODES['reset']

def print_error(text):
    return print_text('darkred', '\n' + text)

def print_message(text, width=80, pad=0, col1='darkgrey', col2='red', footer=1):
    return (
        print_text(col1, '-' * width) + '\n' + 
        print_text(col2, wrap_text(text, width, pad)) + '\n' +
        (footer and print_text(col1, '-' * width))
        )

def print_query(
    text, default, extra='', options=[], width=80, pad=4,
    col1='darkgrey', col2='red'
    ):
    return raw_input(
        print_message('%s ' % text, width, pad, col1, col2, footer='') + 
        extra + 
        print_text(col2, '\n    [') +
        default +
        print_text(col2, '] ')
        ) or default

def print_note(text, col1='green', col2='white', end='\n', start='\n'):
    return print_text(col1, '%s  * ' % start) + print_text(col2, text) + end

# ------------------------------------------------------------------------------
# /dev/null
# ------------------------------------------------------------------------------

class DevNull:
    """Provide a file-like interface emulating /dev/null."""

    def __call__(self, *args, **kwargs):
        pass

    def flush(self):
        pass

    def log(self, *args, **kwargs):
        pass

    def write(self, input):
        pass

DEVNULL = DevNull()

# ------------------------------------------------------------------------------
# Utility Classes
# ------------------------------------------------------------------------------

class IteratorParser:
    """A straightforward parser that lets you move backwards with ease."""

    def __init__(self, sequence):
        self.sequence = iter(sequence)
        self._store = []

    def __iter__(self):
        return self

    def push(self, value):
        self._store.append(value)

    def next(self):
        if self._store:
            return self._store.pop()
        else:
            return self.sequence.next()



# Changes to this file by The Tavutil Authors are in the Public Domain.
# See the Tavutil UNLICENSE file for details.

# This file was adapted from zope/exceptions/exceptionformatter.py in the Zope
# repository and has the following License:

# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.

"""
An exception formatter that shows traceback supplements and traceback info.

  >>> try:
  ...     1/0
  ... except:
  ...     print print_exception()
  Traceback (most recent call last):
    Module pyutil.exception, line 2, in <module>
      1/0
  ZeroDivisionError: integer division or modulo by zero
  <BLANKLINE>

  >>> try:
  ...     1/0
  ... except:
  ...     print print_exception(stream=sys.stdout)
  Traceback (most recent call last):
    Module pyutil.exception, line 2, in <module>
      1/0
  ZeroDivisionError: integer division or modulo by zero
  <BLANKLINE>

Output could be optionally rendered in HTML too:

  >>> try:        # doctest: +NORMALIZE_WHITESPACE
  ...     1/0
  ... except:
  ...     print print_exception(as_html=True)
  <p>Traceback (most recent call last):
  <ul>
  <li>  Module pyutil.exception, line 2, in &lt;module&gt;<br />
      1/0</li>
  </ul>ZeroDivisionError: integer division or modulo by zero<br />
  </p>

"""

import sys
import cgi
import linecache
import traceback

# ------------------------------------------------------------------------------
# some konstants
# ------------------------------------------------------------------------------

__metaclass__ = type

DEBUG_EXCEPTION_FORMATTER = 1

LIMIT = 200

if hasattr(sys, 'tracebacklimit'):
    LIMIT = min(LIMIT, sys.tracebacklimit)

# ------------------------------------------------------------------------------
# the klasses that do the actual formatting
# ------------------------------------------------------------------------------

class TextExceptionFormatter:

    line_sep = '\n'
    show_revisions = 0

    def __init__(self, limit=None, with_filenames=False):
        self.limit = limit
        self.with_filenames = with_filenames

    def escape(self, s):
        return s

    def getPrefix(self):
        return 'Traceback (most recent call last):'

    def getLimit(self):
        limit = self.limit
        if limit is None:
            limit = getattr(sys, 'tracebacklimit', 200)
        return limit

    def formatSupplementLine(self, line):
        return '   - %s' % line

    def formatSourceURL(self, url):
        return [self.formatSupplementLine(url)]

    def formatSupplement(self, supplement, tb):
        result = []
        fmtLine = self.formatSupplementLine

        url = getattr(supplement, 'source_url', None)
        if url is not None:
            result.extend(self.formatSourceURL(url))

        line = getattr(supplement, 'line', 0)
        if line == -1:
            line = tb.tb_lineno
        col = getattr(supplement, 'column', -1)
        if line:
            if col is not None and col >= 0:
                result.append(fmtLine('Line %s, Column %s' % (
                    line, col)))
            else:
                result.append(fmtLine('Line %s' % line))
        elif col is not None and col >= 0:
            result.append(fmtLine('Column %s' % col))

        expr = getattr(supplement, 'expression', None)
        if expr:
            result.append(fmtLine('Expression: %s' % expr))

        warnings = getattr(supplement, 'warnings', None)
        if warnings:
            for warning in warnings:
                result.append(fmtLine('Warning: %s' % warning))

        getInfo = getattr(supplement, 'getInfo', None)
        if getInfo is not None:
            try:
                extra = getInfo()
                if extra:
                    extra = self.escape(extra)
                    if self.line_sep != "\n":
                        extra = extra.replace(" ", "&nbsp;")
                        extra = extra.replace("\n", self.line_sep)
                    result.append(extra)
            except:
                if DEBUG_EXCEPTION_FORMATTER:
                    traceback.print_exc()
                # else just swallow the exception.
        return result

    def formatTracebackInfo(self, tbi):
        return self.formatSupplementLine('__traceback_info__: %s' % (tbi, ))

    def formatLine(self, tb):
        f = tb.tb_frame
        lineno = tb.tb_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        locals = f.f_locals
        globals = f.f_globals

        if self.with_filenames:
            s = '  File "%s", line %d' % (filename, lineno)
        else:
            modname = globals.get('__name__', filename)
            s = '  Module %s, line %d' % (modname, lineno)

        s = s + ', in %s' % name

        result = []
        result.append(self.escape(s))

        # Append the source line, if available
        line = linecache.getline(filename, lineno)
        if line:
            result.append("    " + self.escape(line.strip()))

        # Output a traceback supplement, if any.
        if '__traceback_supplement__' in locals:
            # Use the supplement defined in the function.
            tbs = locals['__traceback_supplement__']
        elif '__traceback_supplement__' in globals:
            # Use the supplement defined in the module.
            # This is used by Scripts (Python).
            tbs = globals['__traceback_supplement__']
        else:
            tbs = None
        if tbs is not None:
            factory = tbs[0]
            args = tbs[1:]
            try:
                supp = factory(*args)
                result.extend(self.formatSupplement(supp, tb))
            except:
                if DEBUG_EXCEPTION_FORMATTER:
                    traceback.print_exc()
                # else just swallow the exception.

        try:
            tbi = locals.get('__traceback_info__', None)
            if tbi is not None:
                result.append(self.formatTracebackInfo(tbi))
        except:
            if DEBUG_EXCEPTION_FORMATTER:
                traceback.print_exc()
            # else just swallow the exception.

        return self.line_sep.join(result)

    def formatExceptionOnly(self, etype, value):
        result = ''.join(traceback.format_exception_only(etype, value))
        return result.replace('\n', self.line_sep)

    def formatLastLine(self, exc_line):
        return self.escape(exc_line)

    def formatException(self, etype, value, tb):
        # The next line provides a way to detect recursion.
        __exception_formatter__ = 1
        result = [self.getPrefix() + '\n']
        limit = self.getLimit()
        n = 0
        while tb is not None and (limit is None or n < limit):
            if tb.tb_frame.f_locals.get('__exception_formatter__'):
                # Stop recursion.
                result.append('(Recursive formatException() stopped)\n')
                break
            line = self.formatLine(tb)
            result.append(line + '\n')
            tb = tb.tb_next
            n = n + 1
        exc_line = self.formatExceptionOnly(etype, value)
        result.append(self.formatLastLine(exc_line))
        return result

class HTMLExceptionFormatter(TextExceptionFormatter):

    line_sep = '<br />\r\n'

    def escape(self, s):
        return cgi.escape(s, 1)

    def getPrefix(self):
        return '<p>Traceback (most recent call last):\r\n<ul>'

    def formatSupplementLine(self, line):
        return '<b>%s</b>' % self.escape(str(line))

    def formatTracebackInfo(self, tbi):
        s = self.escape(str(tbi))
        s = s.replace('\n', self.line_sep)
        return '__traceback_info__: %s' % (s, )

    def formatLine(self, tb):
        line = TextExceptionFormatter.formatLine(self, tb)
        return '<li>%s</li>' % line

    def formatLastLine(self, exc_line):
        return '</ul>%s</p>' % self.escape(exc_line).replace(
            '&lt;br /&gt;',
            '<br />'
            )

# ------------------------------------------------------------------------------
# our kore function
#-------------------------------------------------------------------------------

def format_exception(
    type=None, value=None, traceback=None, limit=LIMIT, as_html=False,
    with_filenames=False
    ):
    """Format a stack trace and the exception information."""

    if not type:
        type, value, traceback = sys.exc_info()

    if as_html:
        formatter = HTMLExceptionFormatter(limit, with_filenames)
    else:
        formatter = TextExceptionFormatter(limit, with_filenames)

    return formatter.formatException(type, value, traceback)

def html_format_exception(type=None, value=None, traceback=None, limit=LIMIT):
    """Format a stack trace and the exception information as HTML."""

    if not type:
        type, value, traceback = sys.exc_info()

    return HTMLExceptionFormatter(limit).formatException(
        type, value, traceback
        )

def print_exception(
    type=None, value=None, traceback=None, limit=LIMIT, as_html=False,
    with_filenames=False, stream=None
    ):
    """Print the exception to the given `stream`."""

    info = format_exception(
        type, value, traceback, limit, as_html, with_filenames
        )

    if stream is None:
        return ''.join(info)
    else:
        for line in info:
            stream.write(line)
        return ''



# Public Domain (-) 2005-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Dictionary Subclasses and Functions."""

# ------------------------------------------------------------------------------
# CachingDict
# ------------------------------------------------------------------------------

Blank = object()

class CachingDict(dict):
    """A caching dict that discards its least recently used items."""

    __slots__ = (
        '_buffer_size', '_cache_size', '_clock', '_garbage_collector'
        )

    def __init__(
        self, cache_size=10000, buffer_size=None, garbage_collector=None, *args,
        **kwargs
        ):

        if buffer_size is None:
            self._buffer_size = cache_size / 2
        else:
            self._buffer_size = buffer_size

        self._cache_size = cache_size
        self._clock = 0
        self._garbage_collector = garbage_collector

        for key, value in args:
            self.__setitem__(key, value)

        for key, value in kwargs.iteritems():
            self.__setitem__(key, value)

    def __setitem__(self, key, value):
        excess = len(self) - self._cache_size - self._buffer_size + 1
        if excess > 0:
            excess = sorted(self.itersort())[:excess + self._buffer_size]
            garbage_collector = self._garbage_collector
            if garbage_collector:
                for ex_value, ex_key in excess:
                    garbage_collector(ex_key, ex_value)
                    del self[ex_key]
            else:
                for _, ex_key in excess:
                    del self[ex_key]
        self._clock += 1
        dict.__setitem__(self, key, [self._clock, value])

    def __getitem__(self, key):
        if key in self:
            access = dict.__getitem__(self, key)
            self._clock += 1
            access[0] = self._clock
            return access[1]
        raise KeyError(key)

    def itersort(self):
        getitem = dict.__getitem__
        for key in self:
            yield getitem(self, key), key

    def get(self, key, default=None):
        if key in self:
            return self.__getitem__(key)
        return default

    def pop(self, key, default=Blank):
        if key in self:
            value = dict.__getitem__(self, key)[1]
            del self[key]
            return value
        if default is not Blank:
            return default
        raise KeyError(key)

    def setdefault(self, key, default):
        if key in self:
            return self.__getitem__(key)
        self.__setitem__(key, default)
        return default

    iterkeys = dict.iterkeys
    keys = dict.keys

    def itervalues(self):
        getitem = self.__getitem__
        for key in self:
            yield getitem(key)

    def values(self):
        getitem = self.__getitem__
        return [getitem(key) for key in self]

    def iteritems(self):
        getitem = self.__getitem__
        for key in self:
            yield key, getitem(key)

    def items(self):
        getitem = self.__getitem__
        return [(key, getitem(key)) for key in self]

    def set_cache_size(self, cache_size, buffer_size=None):
        if buffer_size is None:
            self._buffer_size = cache_size / 2
        else:
            self._buffer_size = buffer_size
        self._cache_size = cache_size

    def get_cache_byte_size(self):
        getitem = self.__getitem__
        return sum(len(str(getitem(key))) for key in self)



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

"""Paypal Adaptive Payments API."""

from demjson import decode as json_decode, encode as json_encode
from urllib2 import urlopen, Request as http_request, URLError

# Generic constants
PRODUCTION_ENDPOINT = 'https://svcs.paypal.com/AdaptivePayments'
SANDBOX_ENDPOINT = 'https://svcs.sandbox.paypal.com/AdaptivePayments'
API_OPERATIONS = ['Pay', 'PaymentDetails', 'Preapproval', \
    'PreapprovalDetails', 'CancelPreapproval', 'ConvertCurrency', 'Refund']


class AdaptivePayment(object):
    """PayPal Adaptive Payments API."""

    def __init__(
            self, api_username, api_password, api_signature, app_id, \
            endpoint=SANDBOX_ENDPOINT
        ):
        self._endpoint = endpoint
        self._api_username = api_username
        self._api_password = api_password
        self._api_signature = api_signature
        self._app_id = app_id

    def call(self, api_operation, api_request):
        try:
            API_OPERATIONS.index(api_operation)
        except ValueError:
            raise RuntimeError("Invalid API operation.  Please refer to"
                      "Adaptive Payments API Operations:"
                      "https://www.x.com/docs/DOC-1408#id099BG0B005Z")
        url = "%s/%s" % (self._endpoint, api_operation)
        req = http_request(url)
        # Authentication
        req.add_header('X-PAYPAL-SECURITY-USERID', self._api_username)
        req.add_header('X-PAYPAL-SECURITY-PASSWORD', self._api_password)
        req.add_header('X-PAYPAL-SECURITY-SIGNATURE', self._api_signature)
        # Specifying data formats
        req.add_header('X-PAYPAL-REQUEST-DATA-FORMAT', 'JSON')
        req.add_header('X-PAYPAL-RESPONSE-DATA-FORMAT', 'JSON')
        # Specifying application information
        req.add_header('X-PAYPAL-APPLICATION-ID', self._app_id)
        json_request = json_encode(api_request)
        try:
            openurl = urlopen(req, json_request)
        except URLError:
            raise RuntimeError("API request failed.")
        return json_decode(openurl.read())


class PayRequest(object):
    """ Creates a PayRequest message which contains the instructions required
    to make a payment.

    Only pass True for receiver_primary when making a chained payment and the
    receiver is actually the designated primary receiver for the transaction.

    You can specify the email address associated with your API credentials in
    the sender_email argument which enables PayPal to implicitly approve a
    payment without the need to redirect to PayPal.
    """

    def __init__(
            self, receiver_email, receiver_amount, currency_code, cancel_url, \
            return_url, receiver_primary=False, sender_email=False,
            preapproval_key=False, pin=False
        ):
        self._pay_request = {}
        self._pay_request['actionType'] = 'PAY'
        self._pay_request['receiverList'] = {'receiver': []}
        self._primary_set = False
        self.add_receiver(rcvr_email=receiver_email,
                rcvr_amount=receiver_amount, rcvr_primary=receiver_primary)
        self._pay_request['currencyCode'] = currency_code
        self._pay_request['cancelUrl'] = cancel_url
        self._pay_request['returnUrl'] = return_url
        self._pay_request['requestEnvelope'] = {'errorLanguage': 'en_US'}

        if sender_email:
            self._pay_request['senderEmail'] = sender_email

        if preapproval_key:
            self._pay_request['preapprovalKey'] = preapproval_key

        if pin:
            self._pay_request['pin'] = pin

    def add_receiver(self, rcvr_email, rcvr_amount, rcvr_primary):
        if len(self._pay_request['receiverList']['receiver']) < 6:
            receiver = {}
            receiver['email'] = rcvr_email
            receiver['amount'] = rcvr_amount

            if self._primary_set:
                if rcvr_primary:
                    raise RuntimeError("PayRequest already has a primary"
                                       "receiver!")
            else:
                if rcvr_primary:
                    receiver['primary'] = 'true'
                    self._primary_set = True

            self._pay_request['receiverList']['receiver'].append(receiver)
        else:
            raise RuntimeError("PayRequest has the maximum six receivers!")

    def get_message(self):
        return self._pay_request


class ResponseEnvelope(object):
    """
    Common response information, including a timestamp and the response
    acknowledgement status.
    """

    def __init__(self, pp_response):
        self._ack = pp_response['responseEnvelope']['ack']
        self._timestamp = pp_response['responseEnvelope']['timestamp']
        self._build = pp_response['responseEnvelope']['build']
        self._correlation_id = pp_response['responseEnvelope']\
                ['correlationId']


class PayResponse(object):
    """
    The PayResponse message which contains a key that you can use to identify
    the payment and the payment's status.
    """

    def __init__(self, pay_response):
        self._response_envelope = ResponseEnvelope( \
               pay_response['responseEnvelope'])
        self._error_data = ErrorData(pay_response['error'])


class PPFault(object):
    """
    The PPFault message returns ErrorData and the ResponseEnvelope
    information to your application if an error occurs when your application
    calls an Adaptive Payments API.
    """

    def __init__(self, pp_fault):
        self._response_envelope = ResponseEnvelope( \
               pp_fault['responseEnvelope'])
        self._error_data = ErrorData(pp_fault['error'])


class ErrorData(object):
    """Detailed error information returned from Paypal API."""

    def __init__(self, error_data):
        for x in error_data:
            self._error[x].category = error_data[x]['category']
            self._error[x].domain = error_data[x]['domain']
            self._error[x].error_id = error_data[x]['errorId']
            self._error[x].message = error_data[x]['message']
            self._error[x].parameter = error_data[x]['parameter']
            self._error[x].severity = error_data[x]['severity']



#! /usr/bin/env python

# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

import sys

from distutils.command.build_ext import build_ext
from setuptools import Extension, setup

# ------------------------------------------------------------------------------
# Extensions
# ------------------------------------------------------------------------------

extensions = [
    Extension(
        "tavutil.lzf",
        ["tavutil/lzf.c", "tavutil/lzf/lzf_c.c", "tavutil/lzf/lzf_d.c"],
        include_dirs=["tavutil/lzf"],
        )
    ]

# TODO(tav): Disable the sandbox for now as it doesn't seem to work on OS X
# Lion.
#
# if sys.platform == 'darwin':
#     extensions.append(
#         Extension("tavutil.darwinsandbox", ["tavutil/darwinsandbox.c"])
#         )

# ------------------------------------------------------------------------------
# Run Setup
# ------------------------------------------------------------------------------

setup(
    name="tavutil",
    author="tav",
    author_email="tav@espians.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: Public Domain",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Programming Language :: Python"
        ],
    cmdclass=dict(build_ext=build_ext),
    description="A collection of utility modules",
    ext_modules=extensions,
    install_requires=[
        "BeautifulSoup==3.2.0",
        "ipaddr>=2.1.7",
        "tornado>=1.2.1"
        ],
    keywords=["async", "crypto", "sandbox", "utility", "redis", "zeroconf"],
    license="Public Domain",
    long_description=open('README.rst').read(),
    packages=["tavutil"],
    url="https://github.com/tav/tavutil",
    version="1.0.1",
    zip_safe=True
    )



# Public Domain (-) 2010-2011 The Tavutil Authors.
# See the Tavutil UNLICENSE file for details.

def wrap_method(method):
    exec("""def wrapper(*args, **kwargs):
        if 'run' in kwargs:
            return method(*args, **kwargs)
        def %s(callback, errback=None):
            kwargs['callback'] = callback
            if errback:
                kwargs['errback'] = errback
            return method(*args, **kwargs)
        return %s""" % (method.__name__, method.__name__), locals())
    wrapper.__name__ = method.__name__
    wrapper.__doc__ = method.__doc__
    wrapper.__raw__ = method
    return wrapper

class Dispatcher(object):
    """An async process dispatcher."""

    def __init__(self, gen):
        self.gen = gen
        self.callback(None)

    def callback(self, arg):
        try:
            self.gen.send(arg)(callback=self.callback, errback=self.errback)
        except StopIteration:
            pass

    def errback(self, arg):
        try:
            self.gen.throw(arg)(callback=self.callback, errback=self.errback)
        except StopIteration:
            pass

def async(func):
    def wrapper(*args, **kwargs):
        Dispatcher(func(*args, **kwargs))
    wrapper.__name__ = func.__name__
    return wrapper
